import datetime
import pytest

from pubtools.pulplib import Repository, Distributor, InvalidDataException


def test_missing_props():
    """from_data raises if input data misses necessary data"""
    with pytest.raises(InvalidDataException):
        Repository.from_data({"missing": "necessary props"})


def test_bad_id():
    """from_data raises if input data has id of wrong type"""
    with pytest.raises(InvalidDataException):
        Repository.from_data({"id": ["foo", "bar", "baz"]})


def test_attr_id():
    """from_data sets id attribute appropriately"""
    repo = Repository.from_data({"id": "some-repo"})
    assert repo.id == "some-repo"


def test_default_created():
    """from_data results in None created by default"""
    repo = Repository.from_data({"id": "some-repo"})
    assert repo.created is None


def test_bad_created(caplog):
    """from_data logs and raises if input data has created of wrong type"""
    with pytest.raises(InvalidDataException):
        Repository.from_data({"id": "some-repo", "notes": {"created": "whoops"}})

    # It should have logged about the bad data. We don't verify details
    # of the failure message since it relies too heavily on implementation
    # details (e.g. stringification of class)
    assert "An error occurred while loading Pulp data!" in caplog.text


def test_is_temporary():
    """from_data is_temporary is True if expected note is present"""
    repo = Repository.from_data({"id": "some-repo", "notes": {"pub_temp_repo": True}})
    assert repo.is_temporary


def test_is_not_temporary():
    """from_data is_temporary is False by default"""
    repo = Repository.from_data({"id": "some-repo"})
    assert not repo.is_temporary


def test_attr_created():
    """from_data sets created attribute appropriately"""
    repo = Repository.from_data(
        {"id": "some-repo", "notes": {"created": "2019-06-11T12:10:00Z"}}
    )

    expected = datetime.datetime(2019, 6, 11, 12, 10, 0, tzinfo=None)
    assert repo.created == expected


def test_distributors_created():
    """from_data sets distributors attribute appropriately"""
    repo = Repository.from_data(
        {
            "id": "some-repo",
            "distributors": [
                {"id": "dist1", "distributor_type_id": "type1"},
                {"id": "dist2", "distributor_type_id": "type1"},
            ],
        }
    )

    assert repo.distributors == [
        Distributor(id="dist1", type_id="type1"),
        Distributor(id="dist2", type_id="type1"),
    ]


def test_distributors_last_publish():
    """from_data sets distributor last publish attribute appropriately"""
    repo = Repository.from_data(
        {
            "id": "some-repo",
            "distributors": [
                {
                    "id": "dist1",
                    "distributor_type_id": "type1",
                    "last_publish": "2019-06-17T01:23:45Z",
                }
            ],
        }
    )

    assert repo.distributor("dist1").last_publish == datetime.datetime(
        2019, 6, 17, 1, 23, 45
    )


def test_distributors_last_publish_null():
    """from_data accepts a null last_publish"""
    repo = Repository.from_data(
        {
            "id": "some-repo",
            "distributors": [
                {"id": "dist1", "distributor_type_id": "type1", "last_publish": None}
            ],
        }
    )

    assert repo.distributor("dist1").last_publish is None


def test_invalid_distributor_repo_id():
    """distributor's repo id being different from repo it's attached is invalid"""

    dist = Distributor(id="dist", type_id="yum_distributor", repo_id="repo")
    with pytest.raises(ValueError) as ex:
        repo = Repository(id="test_repo", distributors=[dist])

    assert "repo_id doesn't match for dist" in str(ex.value)



import datetime
import pytest
import attr

from pubtools.pulplib import FakeController, FileUnit, FileRepository


def test_update_missing_repo():
    controller = FakeController()
    client = controller.client

    # Try to update something which is previously unknown to the client.
    update_f = client.update_repository(FileRepository(id="whatever"))

    # It should fail telling us the repo doesn't exist
    assert "repository not found: whatever" in str(update_f.exception())


def test_can_update_repo():
    controller = FakeController()

    controller.insert_repository(
        FileRepository(id="repo", product_versions=["a", "b", "c"])
    )

    client = controller.client

    # Should be able to get the repo.
    repo = client.get_repository("repo").result()

    # Let's try putting it back. Note we try changing both some mutable
    # and immutable fields here.
    update_f = client.update_repository(
        attr.evolve(repo, eng_product_id=123, product_versions=["d", "b"])
    )

    # The update should succeed (and return None)
    assert update_f.result() is None

    # Try getting the same repo back.
    repo_updated = client.get_repository("repo").result()

    # It should be equal to this:
    assert repo_updated == FileRepository(
        id="repo",
        # product_versions is mutable, so it's what we asked for (with
        # values canonicalized by sorting)
        product_versions=["b", "d"],
        # eng_product_id is not mutable, so that update was ignored
        eng_product_id=None,
    )



import pytest

from pubtools.pulplib import (
    FakeController,
    Criteria,
    RpmUnit,
    ModulemdUnit,
    ModulemdDefaultsUnit,
    YumRepository,
    PulpException,
)


@pytest.fixture
def controller():
    return FakeController()


@pytest.fixture
def populated_units(controller):
    units1 = [
        RpmUnit(
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
        RpmUnit(
            content_type_id="srpm",
            name="bash",
            version="4.0",
            release="1",
            arch="src",
            repository_memberships=["repo1"],
        ),
        RpmUnit(
            name="glibc",
            version="5.0",
            release="1",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
        ModulemdUnit(
            name="module1",
            stream="s1",
            version=1234,
            context="a1b2",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
        ModulemdUnit(
            name="module2",
            stream="s2",
            version=1234,
            context="a1b2",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
        ModulemdDefaultsUnit(
            name="module1", repo_id="repo1", repository_memberships=["repo1"]
        ),
    ]
    units2 = [
        RpmUnit(
            name="glic",
            version="2.3.4",
            release="1",
            arch="x86_64",
            repository_memberships=["repo2"],
        ),
        RpmUnit(
            content_type_id="srpm",
            name="gnu-efi",
            version="3.0c",
            release="1",
            arch="src",
            repository_memberships=["repo2"],
        ),
    ]

    repo1 = YumRepository(id="repo1")
    repo2 = YumRepository(id="repo2")
    controller.insert_repository(repo1)
    controller.insert_units(repo1, units1)
    controller.insert_units(repo2, units2)


def test_search_content_all(populated_units, controller):
    """search_content_by_type with no criteria specified"""
    units1 = [
        u
        for u in controller.client.search_content(
            Criteria.with_field("content_type_id", "rpm")
        ).result()
    ]
    units2 = [
        u
        for u in controller.client.search_content(
            Criteria.with_field("content_type_id", "srpm")
        ).result()
    ]
    units3 = [u for u in controller.client.search_content().result()]
    assert len(units1) == 3
    assert len([u for u in units1 if u.content_type_id == "rpm"]) == 3
    assert len(units2) == 2
    assert len([u for u in units2 if u.content_type_id == "srpm"]) == 2
    # + two modulemd, one modulemd_defaults
    assert len(units3) == 8

    assert set(sum([u.repository_memberships for u in units1], [])) == set(
        ["repo1", "repo2"]
    )


def test_search_content_criteria(populated_units, controller):
    """search_content_by_type with criteria"""
    units1 = [
        u
        for u in controller.client.search_content(
            Criteria.with_field("name", "bash")
        ).result()
    ]
    # match both srpm and rpm
    assert len(units1) == 2
    assert sorted([u.arch for u in units1]) == ["src", "x86_64"]


def test_search_content_invalid_content_type(populated_units, controller):
    """search_content_by_type with invalid content type"""
    with pytest.raises(PulpException):
        for x in controller.client.search_content(
            Criteria.with_field("content_type_id", "invalid")
        ).result():
            pass


def test_search_content_all_pagination(populated_units, controller):
    """search_content_by_type with no criteria specified"""
    units1 = [
        u
        for u in controller.client.search_content(
            Criteria.with_field("content_type_id", "rpm")
        ).result()
    ]
    units2 = [
        u
        for u in controller.client.search_content(
            Criteria.with_field("content_type_id", "srpm")
        ).result()
    ]
    units3 = [u for u in controller.client.search_content().result()]
    assert len(units1) == 3
    assert len([u for u in units1 if u.content_type_id == "rpm"]) == 3
    assert len(units2) == 2
    assert len([u for u in units2 if u.content_type_id == "srpm"]) == 2
    # + two modulemd, one modulemd_defaults
    assert len(units3) == 8

    assert set(sum([u.repository_memberships for u in units1], [])) == set(
        ["repo1", "repo2"]
    )



import sys

import pytest

from pubtools.pulplib import FakeController, FileUnit, FileRepository, PulpException

from ..ioutil import ZeroesIO


def test_can_upload_units(tmpdir):
    """repo.upload_file() succeeds with fake client and populates units."""
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    otherfile = tmpdir.join("another.txt")
    otherfile.write("ahoy there")

    upload1_f = repo1.upload_file(str(somefile))
    upload2_f = repo1.upload_file(str(otherfile), relative_url="another/path.txt")

    for f in [upload1_f, upload2_f]:
        # The future should resolve successfully
        tasks = f.result()

        # The task should be successful.
        assert tasks[0].succeeded

    # If I now search for content in that repo, or content across all repos...
    units_in_repo = sorted(repo1.search_content().result(), key=lambda u: u.sha256sum)
    units_all = sorted(client.search_content().result(), key=lambda u: u.sha256sum)

    # They should be equal
    assert units_all == units_in_repo

    # And they should be this
    assert units_in_repo == [
        FileUnit(
            path="another/path.txt",
            size=10,
            sha256sum="94c0c9d847ecaa45df01999676db772e5cb69cc54e1ff9db31d02385c56a86e1",
            repository_memberships=["repo1"],
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
        ),
        FileUnit(
            path="some-file.txt",
            size=29,
            sha256sum="fad3fc1e6d583b2003ec0a5273702ed8fcc2504271c87c40d9176467ebe218cb",
            repository_memberships=["repo1"],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        ),
    ]


def test_replace_file(tmpdir):
    """repo.upload_file() behaves as expected when replacing a file of the same name."""
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    otherfile = tmpdir.join("another.txt")
    otherfile.write("ahoy there")

    # Upload both files, using the same relative_url for each.
    repo1.upload_file(str(somefile), relative_url="darmok-jalad.txt").result()
    repo1.upload_file(str(otherfile), relative_url="darmok-jalad.txt").result()

    # If I now search for content in that repo, or content across all repos...
    units_in_repo = sorted(repo1.search_content().result(), key=lambda u: u.sha256sum)
    units_all = sorted(client.search_content().result(), key=lambda u: u.sha256sum)

    # I should find that only the second uploaded file is still present in the repo.
    assert units_in_repo == [
        FileUnit(
            path="darmok-jalad.txt",
            size=10,
            sha256sum="94c0c9d847ecaa45df01999676db772e5cb69cc54e1ff9db31d02385c56a86e1",
            repository_memberships=["repo1"],
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
        )
    ]

    # However, both units should still exist in the system; the first uploaded unit
    # has become an orphan.
    assert units_all == units_in_repo + [
        FileUnit(
            path="darmok-jalad.txt",
            size=29,
            sha256sum="fad3fc1e6d583b2003ec0a5273702ed8fcc2504271c87c40d9176467ebe218cb",
            content_type_id="iso",
            repository_memberships=[],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        )
    ]


def test_can_upload_history(tmpdir):
    """repo.upload_file() succeeds with fake client and populates upload_history.

    Note that upload_history is deprecated, but remains working for now.
    """
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    upload_f = repo1.upload_file(str(somefile))

    # The future should resolve successfully
    tasks = upload_f.result()

    # The task should be successful.
    assert tasks[0].succeeded

    # The change should be reflected in the controller's upload history
    history = controller.upload_history

    digest = "fad3fc1e6d583b2003ec0a5273702ed8fcc2504271c87c40d9176467ebe218cb"
    assert len(history) == 1
    assert history[0].repository == repo1
    assert history[0].tasks == tasks
    assert history[0].name == somefile.basename
    assert history[0].sha256 == digest


def test_upload_file_verylarge():
    """Fake client can upload a 2GB file successfully."""
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    file_size = 2000000000
    file_obj = ZeroesIO(file_size)

    upload_f = repo1.upload_file(file_obj, relative_url="big-file")

    # The future should resolve successfully
    tasks = upload_f.result()

    # The task should be successful.
    assert tasks[0].succeeded

    # I should be able to find the corresponding unit.
    units_all = sorted(client.search_content().result(), key=lambda u: u.sha256sum)

    assert units_all == [
        FileUnit(
            path="big-file",
            size=2000000000,
            # If you want to verify this checksum, try:
            #
            #   dd if=/dev/zero bs=1000000 count=2000 status=none | sha256sum
            #
            sha256sum="2e0c654b6cba3a1e816726bae0eac481eb7fd0351633768c3c18392e0f02b619",
            repository_memberships=["repo1"],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        )
    ]


def test_upload_nonexistent_file_raises():
    """repo.upload_file() with nonexistent file fails with fake client"""
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1")

    # If file's not found, Python 2 raises IOError and Python 3 raises
    # FileNotFoundError. The latter one is not defined in Python 2.
    if sys.version_info < (3,):
        exception = IOError
    else:
        exception = FileNotFoundError
    with pytest.raises(exception):
        upload_f = repo1.upload_file("nonexistent_file").result()


def test_upload_repo_absent_raises(tmpdir):
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1")

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    repo_copy1 = client.get_repository("repo1")
    repo_copy2 = client.get_repository("repo1")

    # if repo's deleted
    assert repo_copy1.delete().result()

    exception = repo_copy2.upload_file(str(somefile)).exception()

    assert isinstance(exception, PulpException)
    assert "repo1 not found" in str(exception)



import json
import logging

from pubtools.pulplib import FileRepository, Task, Distributor


def test_publish_update_mappings(fast_poller, requests_mocker, client, caplog):
    """Publish on FileRepository will generate ud_file_release_mappings_2."""

    caplog.set_level(logging.INFO, "pubtools.pulplib")

    repo = FileRepository(
        id="some-repo",
        distributors=[Distributor(id="iso_distributor", type_id="iso_distributor")],
    )
    repo.__dict__["_client"] = client

    # Force client to use a small page size so that we're able to verify
    # all pages end up handled.
    client._PAGE_SIZE = 3

    # Arrange for the repo to currently exist with some mappings.
    requests_mocker.get(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/",
        json={
            "id": "some-repo",
            "notes": {
                "ud_file_release_mappings_2": json.dumps(
                    # Define some mappings so there's a mix of data needing
                    # an update and data needing no update.
                    {
                        "1.0": [{"filename": "file1", "order": 3.0}],
                        "3.0": [{"filename": "file3", "order": 1234}],
                        "4.0": [{"filename": "file4", "order": 0}],
                        "some-other-version": [
                            {"filename": "whatever-file", "order": "abc", "foo": "bar"}
                        ],
                    }
                )
            },
        },
    )

    # Make the repo have some units with the mapping-relevant fields. We use
    # two pages here to ensure that the code implements pagination.
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        [
            {
                # page 1
                "json": [
                    {
                        "metadata": {
                            "_content_type_id": "iso",
                            "name": "file1",
                            "size": 1,
                            "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                            "pulp_user_metadata": {
                                "version": "1.0",
                                "display_order": 3.0,
                            },
                        }
                    },
                    {
                        "metadata": {
                            "_content_type_id": "iso",
                            "name": "file2",
                            "size": 1,
                            "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                        }
                    },
                    {
                        "metadata": {
                            "_content_type_id": "iso",
                            "name": "file3",
                            "size": 1,
                            "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                            "pulp_user_metadata": {"version": "3.0"},
                        }
                    },
                ]
            },
            {
                # page 2
                "json": [
                    {
                        "metadata": {
                            "_content_type_id": "iso",
                            "name": "file4",
                            "size": 1,
                            "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                            "pulp_user_metadata": {
                                "version": "4.0",
                                "display_order": -2,
                            },
                        }
                    },
                    {
                        "metadata": {
                            "_content_type_id": "iso",
                            "name": "file5",
                            "size": 1,
                            "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                            "pulp_user_metadata": {
                                "version": "1.0",
                                "display_order": 4.5,
                            },
                        }
                    },
                    {
                        "metadata": {
                            "_content_type_id": "iso",
                            "name": "file6",
                            "size": 1,
                            "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                            "pulp_user_metadata": {"version": "6.0"},
                        }
                    },
                ]
            },
            {
                # page 3: nothing more
                "json": []
            },
        ],
    )

    # Allow for an update to the repository.
    requests_mocker.put(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/", json={}
    )

    # It should publish as usual.
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        json={"spawned_tasks": [{"task_id": "publish-task"}]},
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[{"task_id": "publish-task", "state": "finished"}],
    )

    # It should have succeeded, with the publish task as retrieved from pulp
    assert sorted(repo.publish()) == [
        Task(id="publish-task", succeeded=True, completed=True)
    ]

    # Now let's have a look at what it updated.
    # The last operations are (update repo, start publish, await tasks),
    # so the update request should always be the 3rd last.
    req = requests_mocker.request_history[-3]

    # Should have been a PUT to the repo.
    assert req.url == "https://pulp.example.com/pulp/api/v2/repositories/some-repo/"
    assert req.method == "PUT"

    # It should have included this field in the delta...
    mapping_json = req.json()["delta"]["notes"]["ud_file_release_mappings_2"]

    # It should have been valid JSON
    mapping = json.loads(mapping_json)

    # It should be equal to this:
    assert mapping == {
        "1.0": [
            # This file wasn't changed
            {"filename": "file1", "order": 3.0},
            # This was added
            {"filename": "file5", "order": 4.5},
        ],
        # This wasn't touched because, although the repo has the file, it has no
        # defined order
        "3.0": [{"filename": "file3", "order": 1234}],
        # This was updated
        "4.0": [{"filename": "file4", "order": -2.0}],
        # This was added, with the entire version not previously defined.
        # Also, order field is omitted because none was in the data, but
        # the version and file still must be written.
        "6.0": [{"filename": "file6"}],
        # This is some unrelated junk which should be left alone during the
        # update process.
        "some-other-version": [
            {"filename": "whatever-file", "foo": "bar", "order": "abc"}
        ],
    }

    # It should have logged that this happened
    assert "Updated ud_file_release_mappings_2 in some-repo" in caplog.messages


def test_publish_update_mappings_noop(fast_poller, requests_mocker, client):
    """Publish on FileRepository does not do unnecessary repo updates."""
    repo = FileRepository(
        id="some-repo",
        distributors=[Distributor(id="iso_distributor", type_id="iso_distributor")],
    )
    repo.__dict__["_client"] = client

    # Arrange for the repo to currently exist with up-to-date mappings.
    requests_mocker.get(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/",
        json={
            "id": "some-repo",
            "notes": {
                "ud_file_release_mappings_2": json.dumps(
                    {
                        "1.0": [
                            {"filename": "file1", "order": 3.0},
                            {"filename": "file2", "order": 6.0},
                        ],
                        "other": ["whatever"],
                    }
                )
            },
        },
    )

    # Make the repo have some units with the mapping-relevant fields.
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        json=[
            {
                "metadata": {
                    "_content_type_id": "iso",
                    "name": "file1",
                    "size": 1,
                    "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                    "pulp_user_metadata": {"version": "1.0", "display_order": 3.0},
                }
            },
            {
                "metadata": {
                    "_content_type_id": "iso",
                    "name": "file2",
                    "size": 1,
                    "checksum": "cd293be6cea034bd45a0352775a219ef5dc7825ce55d1f7dae9762d80ce64411",
                    "pulp_user_metadata": {"version": "1.0"},
                }
            },
        ],
    )

    # It should NOT update the repository, so we don't register any PUT.

    # It should publish as usual.
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        json={"spawned_tasks": [{"task_id": "publish-task"}]},
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[{"task_id": "publish-task", "state": "finished"}],
    )

    # It should have succeeded, with the tasks as retrieved from Pulp
    assert sorted(repo.publish()) == [
        Task(id="publish-task", succeeded=True, completed=True)
    ]



from pubtools.pulplib import ErratumUnit

from pubtools.pulplib._impl.fake.units import is_erratum_version_newer


def test_is_erratum_version_newer():
    """Tests edge cases of internal helper is_erratum_version_newer which are
    cumbersome to verify using public API.
    """

    # Helper to get a unit with a particular version (we don't care about other fields)
    def u(version=None):
        return ErratumUnit(id="whatever", version=version)

    # Empty field on new unit means it's never considered newer (regardless
    # of value on the old unit)
    assert not is_erratum_version_newer(u(), u())
    assert not is_erratum_version_newer(u("0"), u())
    assert not is_erratum_version_newer(u("-3"), u())

    # Old field defaults to 0
    assert not is_erratum_version_newer(u(), u("-1"))
    assert not is_erratum_version_newer(u(), u("0"))
    assert is_erratum_version_newer(u(), u("0.1"))

    # Boring comparisons where everything works normally
    assert is_erratum_version_newer(u("10"), u("20"))
    assert not is_erratum_version_newer(u("20"), u("10"))

    # If *either* field is non-numeric, it's always treated as newer
    assert is_erratum_version_newer(u("oops not a number"), u("10"))
    assert is_erratum_version_newer(u("10"), u("oops not a number"))
    assert is_erratum_version_newer(u("2:30"), u("dentist time"))



import pytest

from pubtools.pulplib._impl.fake.rpmlib import _parse_dep_relation


@pytest.mark.parametrize(
    "flag, expected_relation",
    [
        (0x00, None),
        (0x02, "LT"),
        (0x04, "GT"),
        (0x08, "EQ"),
        (0x0A, "LE"),
        (0x0C, "GE"),
    ],
)
def test_parse_deps_relation(flag, expected_relation):
    """
    Tests expected value of relation according to input flags.
    """
    relation = _parse_dep_relation(flag)

    assert relation == expected_relation



from pubtools.pulplib import FakeController


def test_default_content_type_ids():
    """Fake client by default supports various common types."""
    controller = FakeController()
    client = controller.client

    type_ids = client.get_content_type_ids()

    # Type IDs should include this common content types.
    assert "rpm" in type_ids
    assert "erratum" in type_ids
    assert "modulemd" in type_ids

    # Type IDs reported by client should match controller.
    assert sorted(type_ids) == sorted(controller.content_type_ids)


def test_set_content_type_ids():
    """Fake controller can be used to set content type IDs."""
    controller = FakeController()
    client = controller.client

    controller.set_content_type_ids(["a", "b", "c"])

    type_ids = client.get_content_type_ids()
    assert sorted(type_ids) == ["a", "b", "c"]



import sys
import pytest

from pubtools.pulplib import (
    ModulemdUnit,
    ModulemdDefaultsUnit,
)
from pubtools.pulplib._impl.compat_frozendict import frozendict
from frozenlist2 import frozenlist


def get_test_values():

    default_unit = ModulemdDefaultsUnit(
        unit_id="md_unit_id",
        name="md_unit_name",
        repo_id="md_unit_repo",
        profiles={"1.0": ["default"]},
    )

    md_unit = ModulemdUnit(
        unit_id="unit_id",
        name="name",
        stream="1.0",
        version=1,
        context="context",
        arch="x86_64",
        profiles={
            "default": {
                "description": "Default description",
                "rpms": ["rpm"],
            },
        },
    )
    return default_unit, md_unit


@pytest.mark.xfail(
    sys.version_info < (3, 0),
    reason="The conversion function checks if the value is already a frozendict, which is just "
    "dict in python2, so profiles and the contents wont be converted",
)
def test_convert_profile_value():
    """Profile and the values it contains should be immutable"""

    default_unit, md_unit = get_test_values()

    assert isinstance(default_unit.profiles, frozendict)
    assert isinstance(default_unit.profiles["1.0"], frozenlist)
    assert isinstance(md_unit.profiles, frozendict)
    assert isinstance(md_unit.profiles["default"], frozendict)
    assert isinstance(md_unit.profiles["default"]["rpms"], frozenlist)


@pytest.mark.xfail(
    sys.version_info < (3, 0),
    reason="Frozendict falls back to the default dict for Python2, so wont have a hash",
)
def test_profiles_have_hash():
    """If two objects are equal, then they must have the same hashcode"""
    default_unit_1, md_unit_1 = get_test_values()
    default_unit_2, md_unit_2 = get_test_values()

    assert hash(default_unit_1) == hash(default_unit_2)
    assert hash(md_unit_1) == hash(md_unit_2)


@pytest.mark.xfail(
    sys.version_info < (3, 0),
    reason="Frozendict falls back to the default dict for Python2, so wont throw exceptions",
)
def test_fail_on_modify_profiles():
    """Should fail when the profile attribute is modified"""
    default_unit, md_unit = get_test_values()
    with pytest.raises(AttributeError):
        default_unit.profiles.clear()

    with pytest.raises(TypeError):
        default_unit.profiles["1.0"] = ["modify"]

    with pytest.raises(NotImplementedError):
        default_unit.profiles["1.0"].append("modify")

    with pytest.raises(TypeError):
        md_unit.profiles["default"] = {}

    with pytest.raises(TypeError):
        md_unit.profiles["default"]["description"] = "modify"



import json
import os

import pytest

from pubtools.pulplib import (
    FakeController,
    Repository,
    FileRepository,
    MaintenanceReport,
)

MaintenanceReport._OWNER = "ContentDelivery"


def test_set_maintenance():

    controller = FakeController()

    maintain_repo = FileRepository(id="redhat-maintenance")

    controller.insert_repository(maintain_repo)
    # now the maintenance repo is empty

    client = controller.client

    report = client.get_maintenance_report().result()

    assert isinstance(report, MaintenanceReport)
    # return an empty report
    assert report.entries == []
    # there's no entries in the report

    report = report.add(repo_ids=["repo1", "repo2"])
    client.set_maintenance(report).result()
    # add entries to report and set maintenance

    # upload_file and publish should be called once each
    assert len(controller.upload_history) == 1
    assert len(controller.publish_history) == 1

    # get_maintenance_report should give a report object
    report = client.get_maintenance_report().result()
    assert report.last_updated_by == "ContentDelivery"
    assert len(report.entries) == 2

    report = report.remove(repo_ids=["repo1"], owner="jazhang@hostname")
    client.set_maintenance(report).result()

    # the report in repo should have updated
    report = client.get_maintenance_report().result()
    assert report.last_updated_by == "jazhang@hostname"
    assert len(report.entries) == 1
    assert report.entries[0].repo_id == "repo2"



import pytest

from pubtools.pulplib import Repository, Task, DetachedException


def test_detached():
    """delete raises if called on a detached repo"""
    with pytest.raises(DetachedException):
        Repository(id="some-repo").delete()


def test_delete_success(fast_poller, requests_mocker, client):
    """delete succeeds and returns spawned tasks"""

    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    requests_mocker.delete(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/",
        json={"spawned_tasks": [{"task_id": "task1"}, {"task_id": "task2"}]},
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[
            {"task_id": "task1", "state": "finished"},
            {"task_id": "task2", "state": "skipped"},
        ],
    )

    # It should have succeeded, with the tasks as retrieved from Pulp
    assert sorted(repo.delete()) == [
        Task(id="task1", succeeded=True, completed=True),
        Task(id="task2", succeeded=True, completed=True),
    ]


def test_delete_detaches(fast_poller, requests_mocker, client):
    """delete causes repository to become detached"""

    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    requests_mocker.delete(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/",
        json={"spawned_tasks": [{"task_id": "task1"}, {"task_id": "task2"}]},
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[
            {"task_id": "task1", "state": "finished"},
            {"task_id": "task2", "state": "skipped"},
        ],
    )

    # Delete should succeed
    repo.delete().result()

    # Second attempt to delete doesn't make sense
    with pytest.raises(DetachedException):
        repo.delete()


def test_delete_absent(fast_poller, requests_mocker, client):
    """delete of an object which already doesn't exist is successful"""

    # Two handles to same repo
    repo1 = Repository(id="some-repo")
    repo1.__dict__["_client"] = client

    repo2 = Repository(id="some-repo")
    repo2.__dict__["_client"] = client

    requests_mocker.delete(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/",
        [
            # first attempt to delete
            dict(json={"spawned_tasks": [{"task_id": "task1"}]}),
            # second attempt fails as 404 because it already was deleted
            dict(
                status_code=404,
                json={"http_status": 404, "http_request_method": "DELETE"},
            ),
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[{"task_id": "task1", "state": "finished"}],
    )

    # Delete via first handle succeeds with the spawned task
    assert sorted(repo1.delete()) == [Task(id="task1", succeeded=True, completed=True)]

    # Delete via second handle also succeeds, but with no tasks
    assert sorted(repo2.delete()) == []



import pytest
from pubtools.pulplib import Repository, DetachedException, Criteria, Matcher, FileUnit


def test_detached():
    """content searches raise if called on a detached repository object"""
    with pytest.raises(DetachedException):
        assert not Repository(id="some-repo").file_content


def test_complex_type_ids(client):
    """content searches raise if using criteria with unsupported operators on content_type_id"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    with pytest.raises(ValueError) as e:
        repo.search_content(
            Criteria.with_field("content_type_id", Matcher.regex("foobar"))
        )

    assert "unsupported expression for content_type_id" in str(e.value)


def test_mixed_search(client, requests_mocker):
    """Searching with a criteria mixing several fields works correctly"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        json=[
            {
                "metadata": {
                    "_content_type_id": "iso",
                    "name": "hello.txt",
                    "size": 23,
                    "checksum": "a" * 64,
                }
            }
        ],
    )

    crit = Criteria.and_(
        Criteria.with_field_in("content_type_id", ["rpm", "iso"]),
        Criteria.with_field("name", "hello.txt"),
    )

    files = list(repo.search_content(crit))

    assert files == [FileUnit(path="hello.txt", size=23, sha256sum="a" * 64)]

    history = requests_mocker.request_history

    # There should have been just one request
    assert len(history) == 1

    request = history[0]
    body = request.json()

    # This should have been the request body.
    assert body == {
        "criteria": {
            "type_ids": ["rpm", "iso"],
            "skip": 0,
            "limit": 2000,
            "filters": {"unit": {"name": {"$eq": "hello.txt"}}},
        }
    }


def test_search_content_type_id_in_or(client, requests_mocker):
    """Searching with a content_type_id within $or fails as unsupported"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    crit = Criteria.or_(
        Criteria.with_field("name", "hello.txt"),
        Criteria.with_field_in("content_type_id", ["rpm", "iso"]),
    )

    with pytest.raises(ValueError) as e:
        repo.search_content(crit).result()

    assert "Can't serialize criteria for Pulp query; too complicated" in str(e.value)


def test_file_content(client, requests_mocker):
    """file_content returns correct unit types"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        json=[
            {
                "metadata": {
                    "_content_type_id": "iso",
                    "name": "hello.txt",
                    "size": 23,
                    "checksum": "a" * 64,
                }
            }
        ],
    )

    files = repo.file_content

    assert len(files) == 1
    assert files[0].content_type_id == "iso"


def test_rpm_content(client, requests_mocker):
    """rpm_content returns correct unit types"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        json=[
            {
                "metadata": {
                    "_content_type_id": "rpm",
                    "name": "bash",
                    "epoch": "0",
                    "filename": "bash-x86_64.rpm",
                    "version": "4.0",
                    "release": "1",
                    "arch": "x86_64",
                }
            }
        ],
    )

    rpms = repo.rpm_content

    assert len(rpms) == 1
    assert rpms[0].content_type_id == "rpm"


def test_srpm_content(client, requests_mocker):
    """srpm_content returns correct unit types"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        json=[
            {
                "metadata": {
                    "_content_type_id": "srpm",
                    "name": "bash",
                    "epoch": "0",
                    "filename": "bash-x86_64.srpm",
                    "version": "4.0",
                    "release": "1",
                    "arch": "x86_64",
                }
            }
        ],
    )

    srpms = repo.srpm_content

    assert len(srpms) == 1
    assert srpms[0].content_type_id == "srpm"


def test_modulemd_content(client, requests_mocker):
    """modulemd_content returns correct unit types"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        json=[
            {
                "metadata": {
                    "_content_type_id": "modulemd",
                    "name": "md",
                    "stream": "s1",
                    "version": 1234,
                    "context": "a1b2c3",
                    "arch": "s390x",
                }
            }
        ],
    )

    modulemds = repo.modulemd_content

    assert len(modulemds) == 1
    assert modulemds[0].content_type_id == "modulemd"


def test_modulemd_defaults_content(client, requests_mocker):
    """modulemd_defaults_content returns correct unit types"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/search/units/",
        json=[
            {
                "metadata": {
                    "_content_type_id": "modulemd_defaults",
                    "name": "mdd",
                    "repo_id": "some-repo",
                    "profiles": {"p1": ["something"]},
                }
            }
        ],
    )

    modulemd_defaults = repo.modulemd_defaults_content

    assert len(modulemd_defaults) == 1
    assert modulemd_defaults[0].content_type_id == "modulemd_defaults"



from functools import partial

import attr
import pytest

from pubtools.pulplib import FakeController, Repository, PulpException, FileUnit


def test_can_construct():
    """A fake client can be constructed."""
    controller = FakeController()
    assert controller.client is not None


def test_can_get():
    """get_repository returns repository inserted via controller."""
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)

    client = controller.client
    found = client.get_repository("repo2").result()

    assert found == repo2


def test_get_missing_raises():
    """get_repository returns unsuccessful future if repo doesn't exist."""
    controller = FakeController()

    repo_f = controller.client.get_repository("some-repo")
    with pytest.raises(PulpException) as raised:
        repo_f.result()

    assert "some-repo not found" in str(raised.value)


def test_get_wrong_type_raises():
    """get_repository raises TypeError if passed argument of wrong type."""
    controller = FakeController()

    client = controller.client
    with pytest.raises(TypeError):
        client.get_repository(["oops", "should have been a string"])


def test_client_lifecycle():
    """FakeClient can be used in a with statement, and not afterwards."""
    controller = FakeController()

    with controller.client as client:
        # This should work OK
        assert client.search_repository().result()

    # But after end of 'with' statement, most public methods will no longer work
    for fn in [
        client.search_repository,
        client.search_content,
        client.search_distributor,
        partial(client.get_repository, "somerepo"),
        client.get_maintenance_report,
        partial(client.set_maintenance, {"what": "ever"}),
        client.get_content_type_ids,
    ]:
        with pytest.raises(RuntimeError) as excinfo:
            fn()

        assert "cannot schedule new futures after shutdown" in str(excinfo.value)


def test_can_insert_orphans():
    """insert_units with a null repo inserts units as orphans."""
    controller = FakeController()

    units = [
        FileUnit(
            path="bar",
            size=0,
            sha256sum="b1a6cb41223dcd02f208827517fe4b59a12684b76e15dbee645e9f9a9daa952e",
        ),
        FileUnit(
            path="quux",
            size=0,
            sha256sum="b1a6cb41223dcd02f208827517fe4b59a12684b76e15dbee645e9f9a9daa952e",
        ),
    ]

    # Can insert
    controller.insert_units(None, units)

    client = controller.client

    # If I now search for all content...
    found = list(client.search_content())

    for unit in found:
        # It should have an ID
        assert unit.unit_id
        # It should have memberships [] (no repos) rather than None (unknown repos)
        assert unit.repository_memberships == []

    # Other than those two fields, it should be identical to the input
    found_cmp = [
        attr.evolve(u, unit_id=None, repository_memberships=None) for u in found
    ]
    assert sorted(found_cmp, key=repr) == units
